/*****************************************************************************/
/* XKeyWrap - A simple keyboard wrapper on X Window System.                  */
/*                                                                           */
/* XKeyWrap Copyright (C) 1999 Sakai Hiroaki.                                */
/* All Rights Reserved.                                                      */
/*****************************************************************************/

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include "list.h"

typedef struct _ListData {
  char action;
  KeySym key_sym;
  unsigned int state;
  Time timer;
  struct _ListData * next;
} _ListData;

typedef struct _List {
  _ListData * start;
  _ListData * end;
} _List;

static _ListData * AllocListData(char action, KeySym key_sym,
				 unsigned int state, Time timer)
{
  _ListData * p;
  p = (_ListData *)malloc(sizeof(_ListData));
  p->action = action;
  p->key_sym = key_sym;
  p->state = state;
  p->timer = timer;
  p->next = NULL;
  return (p);
}

ListPointer ListGetStart(List this)
{
  return (this->start);
}

ListPointer ListGetEnd(List this)
{
  return (this->end);
}

ListPointer ListGetNext(ListPointer pointer)
{
  return (pointer->next);
}

char ListGetAction(ListPointer pointer)
{
  return (pointer->action);
}

KeySym ListGetKey(ListPointer pointer)
{
  return (pointer->key_sym);
}

unsigned int ListGetState(ListPointer pointer)
{
  return (pointer->state);
}

Time ListGetTimer(ListPointer pointer)
{
  return (pointer->timer);
}

ListPointer ListAddToStart(List this, char action, KeySym key_sym,
			   unsigned int state, Time timer)
{
  _ListData * p;

  p = AllocListData(action, key_sym, state, timer);
  if (this->start == NULL) {
    this->start = p;
    this->end   = p;
  } else {
    p->next = this->start;
    this->start = p;
  }
  return (p);
}

ListPointer ListAddToEnd(List this, char action, KeySym key_sym,
			 unsigned int state, Time timer)
{
  _ListData * p;

  p = AllocListData(action, key_sym, state, timer);
  if (this->start == NULL) {
    this->start = p;
    this->end   = p;
  } else {
    this->end->next = p;
    this->end = p;
  }
  return (p);
}

ListPointer ListAdd(List this, ListPointer pointer,
		    char action, KeySym key_sym,
		    unsigned int state, Time timer)
{
  _ListData * p;

  if (pointer == NULL) {
    return (ListAddToStart(this, action, key_sym, state, timer));
  } else if (pointer->next == NULL) {
    return (ListAddToEnd(this, action, key_sym, state, timer));
  } else {
    p = AllocListData(action, key_sym, state, timer);
    p->next = pointer->next;
    pointer->next = p;
  }
  return (p);
}

void ListClear(List this)
{
  _ListData * p;
  _ListData * n;

  if (this == NULL) return;

  p = this->start;
  while (p) {
    n = p->next;
    free(p);
    p = n;
  }

  this->start = NULL;
  this->end = NULL;
}

List ListCreate()
{
  List list;
  list = (List)malloc(sizeof(_List));
  list->start = NULL;
  list->end   = NULL;
  return (list);
}

void ListDestroy(List this)
{
  _ListData * p;
  _ListData * n;

  if (this == NULL) return;

  p = this->start;
  while (p) {
    n = p->next;
    free(p);
    p = n;
  }

  free(this);
}

/*****************************************************************************/
/* End of file.                                                              */
/*****************************************************************************/
