/*****************************************************************************/
/* XKeyWrap - A simple keyboard wrapper on X Window System.                  */
/*                                                                           */
/* XKeyWrap Copyright (C) 1999-2001 Sakai Hiroaki.                           */
/* All Rights Reserved.                                                      */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>

#include "list.h"
#include "readme.h"
#include "copyright.h"

#define DEFAULT_FILENAME "keyseq.dat"

/* ˥塼ΥեȤΥ */
#define FONT_WIDTH   8
#define FONT_HEIGHT 16
#define FONT_OFFSET 14

#define WIDTH  (FONT_WIDTH  * menubar_len)
#define HEIGHT (FONT_HEIGHT * 2)

/* X Ϣ */
static Widget toplevel = NULL;           /* ȥåץ٥Υå */
static Display * display = NULL;         /* ǥץ쥤 */
static Window root_window = 0;           /* 롼ȥɥ */

/* ΥɥϢ */
static Display * output_display = NULL;  /* ٥Ƚǥץ쥤 */
static Window output_root_window = 0;    /* ٥Ƚ롼ȥɥ */
static Window output_window = 0;         /* ٥Ƚ襦ɥ */
static Cursor select_cursor = NULL;      /* 襦ɥΥ */
static int cursor_defined = 0;           /* /̤Υե饰 */

static char ** child_argv = NULL;        /* ҥץ */
static int child = 0;                    /* fork() ҥץ */
static char * filename = NULL;           /* ǡեΥե̾ */
static Font font = 0;                    /* ˥塼ɽѤΥե */
static GC font_gc = 0;                   /* ʸɽѤ GC */
static int focus_menu = 0;               /* եƤ˥塼 */
static List key_list = NULL;             /* 󥹤εϿѥꥹ */
static int recording = 0;                /* Ͽ椫ɤ */
static Time recording_timer = 0;         /* Ͽ˳ϻ֤򵭲 */
static int playing = 0;                  /* 椫ɤ */
static int repeating = 0;                /* ֤뤫ɤ */
static int pausing = 0;                  /* ߾֤Υե饰 */
static ListPointer timer_pointer = NULL; /* Play ˥󥹤ؤ */
static XtIntervalId timeout_id = 0;      /* Play Υॢ */
static int current_number = 0;           /* ҥɥ */
static int children_number = 0;          /* ҥɥ */
static Window * children_list = NULL;    /* ҥɥ */
static Atom wm_delete_window;            /* f.delete ؤб */
static char menubar[200];                /* ˥塼Сɽ */
static int menubar_len;                  /* ˥塼Сɽ */
static int auto_quit = 0;                /* ҥץλ˽λ */

/* Play ® */
static int play_speed = 5;
static int speed[] = {0, 1, 2, 4, 8, 16, 32, 64, 128, 256};
#define SPEED_NUM_MAX (sizeof(speed) / sizeof(speed[0]))

static void wm_quit(Widget w, XEvent * ev, String * params, Cardinal * num);

/*  */
static XtActionsRec actions[] = {
  {"wm_quit", wm_quit}
};

/* ȥ󥹥졼 */
static char * translations =
"<Message>WM_PROTOCOLS: wm_quit()";

/* ˥塼 */
static void select_window(XEvent * ev);
static void search_parent(XEvent * ev);
static void search_child(XEvent * ev);
static void next_child(XEvent * ev);
static void clear_sequence(XEvent * ev);
static void record_sequence(XEvent * ev);
static void stop_recording(XEvent * ev);
static void play_sequence(XEvent * ev);
static void pause_sequence(XEvent * ev);
static void repeat_sequence(XEvent * ev);
static void speed_up(XEvent * ev);
static void speed_down(XEvent * ev);
static void save_sequence(XEvent * ev);
static void load_sequence(XEvent * ev);
static void quit(XEvent * ev);

static struct {
  char * name;
  void (*func)(XEvent *);
  char * information;
} menu_struct[] = {
  {"Select", select_window  ,
   "Select the window to output key sequence. Click left button on it."},
  {"Parent", search_parent  ,
   "Change the window to output key sequence to parent."},
  {"Child" , search_child   ,
   "Change the window to output key sequence to one of children."},
  {"Next"  , next_child     ,
   "Change the window to output key sequence to one of other children."},
  {"Clear" , clear_sequence , "Clear the recorded key sequence."},
  {"Record", record_sequence,
   "Start recording key sequence. Double click to record to end of list."},
  {"Stop"  , stop_recording , "Stop recording and playing."},
  {"Play"  , play_sequence  , "Play the recorded key sequence."},
  {"Pause" , pause_sequence ,
   "Pause or restart, recording or playing key sequence."},
  {"Repeat", repeat_sequence, "Auto repeat playing the recorded key."},
  {"+"     , speed_up       , "Change play speed faster."},
  {"-"     , speed_down     , "Change play speed more slow."},
  {"Save"  , save_sequence  , "Save the key sequence data to file."},
  {"Load"  , load_sequence  , "Load the key sequence data from file."},
  {"Quit"  , quit           , "Exit XKeyWrap."},
  {NULL, NULL, NULL}
};

static char * help_message =
"\n" VERSION "\n\n"
"XKeyWrap - A simple keyboard wrapper on X Window System.\n"
"\n"
"Usage :\n"
"xkeywrap [-h, -help] [-readme] [-copyright]\n"
"         [-p, -play] [-r, -repeat] [-rec, -record] [-pause] [-q, -quit]\n"
"         [-s, -speed n] [-f, -file filename]\n"
"         [-output-display displayname]\n"
"         [-output-window windowid] [-display displayname] [-geometry +X+Y]\n"
"         [-iconic] [-bg color] [-name name]\n"
"         [-x Application [Application's options...]]\n"
"\n"
"If you do not understand about using XKeyWrap, type\n"
"\n"
"    % xkeywrap -readme\n"
"\n"
"for more informations and examples about using XKeyWrap.\n"
"\n";

/*****************************************************************************/
/* ޤޤʴؿ                                                            */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* åν(ʸΤ)                                              */
/*---------------------------------------------------------------------------*/

static void output_message(char * message)
{
  static char buffer[200];
  char buf2[10];
  int i, x, len;

  if (message) strncpy(buffer, message, 200);

  XClearWindow(display, XtWindow(toplevel));
  XDrawString(display, XtWindow(toplevel), font_gc,
	      0, FONT_OFFSET, menubar, menubar_len);

  x = 0;
  for (i = 0; menu_struct[i].name; i++) {
    len = strlen(menu_struct[i].name);
    /*
      XDrawString(display, XtWindow(toplevel), font_gc,
      x * FONT_WIDTH, FONT_OFFSET, menu_struct[i].name, len);
      */
    if (i == focus_menu)
      XDrawRectangle(display, XtWindow(toplevel), font_gc,
		     x * FONT_WIDTH, 0, len * FONT_WIDTH - 1, FONT_HEIGHT - 1);
    x += len + 1;
  }
  XDrawString(display, XtWindow(toplevel), font_gc,
	      0, FONT_HEIGHT + FONT_OFFSET, buffer, strlen(buffer));

  sprintf(buf2, "%c%c%c%d",
	  (recording ? 'R' : ' '),
	  (repeating ? 'A' : ' '),
	  (pausing   ? 'P' : ' '),
	  (playing % 10));

  XDrawString(display, XtWindow(toplevel), font_gc,
	      FONT_WIDTH * (menubar_len - 4), FONT_HEIGHT + FONT_OFFSET,
	      buf2, 4);

  sprintf(buf2, "%3d/16", speed[play_speed]);
  XDrawString(display, XtWindow(toplevel), font_gc,
	      FONT_WIDTH * (menubar_len - 10), FONT_HEIGHT + FONT_OFFSET,
	      buf2, 6);

  XFlush(display);
}

/*---------------------------------------------------------------------------*/
/* åν()                                                */
/*---------------------------------------------------------------------------*/

static void output_value(char * message, int value)
{
  static char buffer[200];
  sprintf(buffer, message, value);
  output_message(buffer);
}

/*---------------------------------------------------------------------------*/
/* åν(ʸ)                                              */
/*---------------------------------------------------------------------------*/

static void output_string(char * message, char * string)
{
  static char buffer[200];
  sprintf(buffer, message, string);
  output_message(buffer);
}

/*---------------------------------------------------------------------------*/
/* XQueryTree() ҥɥĴ٤                           */
/*---------------------------------------------------------------------------*/

Window diff_windows(Window * from, unsigned int n_from,
		    Window * to, unsigned int n_to)
{
  int i, j;
  int f;

  for (i = 0; i < n_to; i++) {
    f = 0;
    for (j = 0; j < n_from; j++) {
      if (to[i] == from[j]) {
	f = 1;
	break;
      }
    }
    if (!f) return (to[i]);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ॢȤλ֤׻                                              */
/*---------------------------------------------------------------------------*/

int cal_timeout(Time timer, int s)
{
  int t;

  t = timer * speed[s] / 16;
  if (t < 1) return (1);
  return (t);
}

/*****************************************************************************/
/* ˥塼                                                                  */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* ޥݥ󥿤ư顤˥塼Υեư                */
/*---------------------------------------------------------------------------*/

static void pointer_motion(Widget w, XtPointer client_data, XEvent * ev,
			   Boolean * dispatch)
{
  int i, x;
  static int old_menu = -1;

  if (ev->type == MotionNotify) {
    if (ev->xmotion.y <= FONT_HEIGHT) {
      x = 0;
      for (i = 0; menu_struct[i].name; i++) {
	x += strlen(menu_struct[i].name) + 1;
	if (ev->xmotion.x <= x * FONT_WIDTH) {
	  focus_menu = i;
	  break;
	}
      }
      if (old_menu != focus_menu)
	output_message(menu_struct[focus_menu].information);
      old_menu = focus_menu;
    }
  }
}

/*---------------------------------------------------------------------------*/
/* ޥΥܥ򥯥å顤եƤ˥塼              */
/* ˥塼ϥɥ¹Ԥ                                                */
/*---------------------------------------------------------------------------*/

static void button_press(Widget w, XtPointer client_data, XEvent * ev,
			 Boolean * dispatch)
{
  if (ev->xbutton.y <= FONT_HEIGHT) {
    (*(menu_struct[focus_menu].func))(ev);
  }
}

/*===========================================================================*/
/* ˥塼Υϥɥؿ                                                    */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ɥ                                                          */
/*---------------------------------------------------------------------------*/

static void select_window(XEvent * ev)
{
  Window r, child, child2;
  int rx, ry, wx, wy;
  unsigned int mask;
  int counter;

  output_message("Click left button on a window to send key sequence.");

  /* ޥκܥ󤬤äΥޤԤ */
  counter = 0;
  do {
    XQueryPointer(output_display, output_root_window,
		  &r, &child, &rx, &ry, &wx, &wy, &mask);
    if (counter++ > 100) {
      output_message(NULL);
      counter = 0;
    }
    usleep(100);
  } while (mask & Button1Mask);

  /* ޥѹ */
  XDefineCursor(output_display, output_root_window, select_cursor);
  cursor_defined = 1;
  XFlush(output_display);

  /* ޥκܥ󤬲ޤԤ */
  counter = 0;
  do {
    XQueryPointer(output_display, output_root_window,
		  &r, &child, &rx, &ry, &wx, &wy, &mask);
    if (counter++ > 100) {
      output_message(NULL);
      counter = 0;
    }
    usleep(100);
  } while (!(mask & Button1Mask));
  if (child == 0) child = output_root_window;

  /*
   * ɥ˵٥Ȥ˴롥
   * ʤȡ㤨 Select ˥塼򤷤 Parent ˥塼
   * Ȥ̤᤮Ƥ̤Υɥ򤷤ݤˡޥưΥ٥Ȥ
   * ĤäƤޤɥIDɽľ Parent ˥塼ФƤޤ
   * ɥIDʤȤݤ롥
   */
  XSync(display, True);

  /* ޥ᤹ */
  XUndefineCursor(output_display, output_root_window);
  cursor_defined = 0;
  XFlush(output_display);

  /* ޥݥ󥿤ޤҥɥ򸡺 */
  while (1) {
    XQueryPointer(output_display, child,
		  &r, &child2, &rx, &ry, &wx, &wy, &mask);
    if (child2 == 0) break;
    child = child2;
  }

  output_window = child;

  output_value("Window ID %d is selected.", (int)output_window);
}

/*---------------------------------------------------------------------------*/
/* ƥɥθ                                                        */
/*---------------------------------------------------------------------------*/

static void search_parent(XEvent * ev)
{
  Window r, p;
  Window * children;
  unsigned int n;

  if (output_window == output_root_window) {
    output_message("This is root window.");
    return;
  }

  XQueryTree(output_display, output_window, &r, &p, &children, &n);
  XFree(children);
  output_window = p;

  if (output_window == output_root_window)
    output_message("Root window.");
  else
    output_value("Window id is %d.", (int)output_window);

  if (children_list) XFree(children_list);
  children_number = 0;
  current_number = 0;
  children_list = NULL;
}

/*---------------------------------------------------------------------------*/
/* ҥɥθ                                                        */
/*---------------------------------------------------------------------------*/

static void search_child(XEvent * ev)
{
  Window r, p;

  if (children_list) XFree(children_list);

  XQueryTree(output_display, output_window,
	     &r, &p, &children_list, &children_number);

  if (!children_number) {
    output_message("No child windows.");
  } else {
    output_window = children_list[current_number = 0];
    output_value("Window id is %d.", (int)output_window);
  }
}

/*---------------------------------------------------------------------------*/
/* λҥɥθ                                                    */
/*---------------------------------------------------------------------------*/

static void next_child(XEvent * ev)
{
  if (!children_number) {
    output_message("No child windows.");
  } else {
    current_number =
      current_number < children_number - 1 ? current_number + 1 : 0;
    output_window = children_list[current_number];
    output_value("Window id is %d.", (int)output_window);
  }
}

/*---------------------------------------------------------------------------*/
/* ꥹȤν                                                            */
/*---------------------------------------------------------------------------*/

static void all_stop()
{
  recording = 0;
  playing = 0;
  pausing = 0;
  timer_pointer = NULL;
  if (timeout_id) XtRemoveTimeOut(timeout_id);
}

static void clear_sequence(XEvent * ev)
{
  all_stop();
  ListClear(key_list);
  output_message("Clear all key sequence. OK.");
}

/*---------------------------------------------------------------------------*/
/* 쥳ǥ󥰤γ                                                      */
/*---------------------------------------------------------------------------*/

static void record_sequence(XEvent * ev)
{
  if (recording) {
    timer_pointer = ListGetEnd(key_list);
    output_message("Start recording from end of key list.");
  } else {
    recording = 1;
    recording_timer = 0;
    if (ev) {
      if (ev->type == ButtonPress) {
	recording_timer = ev->xbutton.time;
      }
    }
#if 0
    if (playing && pausing) playing = 0;
#endif
    output_message("Start recording.");
  }
}

/*---------------------------------------------------------------------------*/
/* 쥳ǥ󥰡ꥢ륿                                    */
/*---------------------------------------------------------------------------*/

static void stop_recording(XEvent * ev)
{
  if (!pausing) {
    all_stop();
  }
  playing = 0;
  recording = 0;

  output_message("Stopped.");
}

/*---------------------------------------------------------------------------*/
/* 쥳ǥ󥰤ǡκ(ꥢ륿)ΥॢѴؿ      */
/*---------------------------------------------------------------------------*/

static void play_proc(XtPointer p, XtIntervalId * id);

static void start_playing()
{
  if (ListGetStart(key_list)) {
    timeout_id =
      XtAddTimeOut(cal_timeout(ListGetTimer(ListGetStart(key_list)),
			       play_speed),
		   play_proc, NULL);
    output_message("Start playing.");
  } else {
    playing = 0;
    timeout_id = 0;
    output_message("No data.");
  }
}

static void finish_playing()
{
  if (repeating) { /* ֤ */
    output_message("Finished playing.");
    start_playing();
  } else {
    if (playing > 0) playing--;
    /* playing--; Ƥ output_message() ɬפ */
    output_message("Finished playing.");
    if (playing) {
      start_playing();
    } else {
      timeout_id = 0;
      timer_pointer = NULL;
    }
  }
}

static void restart_playing()
{
  int t;
  if (timer_pointer) {
    if (ListGetNext(timer_pointer)) {
      t = cal_timeout(ListGetTimer(ListGetNext(timer_pointer)), play_speed);
      timeout_id = XtAddTimeOut(t, play_proc, NULL);
      output_message("Restart playing.");
      return;
    }
  }
  start_playing();
}

static void play_proc(XtPointer p, XtIntervalId * id)
{
  static XEvent event;
  int t;

  if (!output_window) {
    playing = 0;
    timeout_id = 0;
    timer_pointer = NULL;
    output_message("No output window.");
    return;
  }

  if (!timer_pointer) timer_pointer = ListGetStart(key_list);
  else timer_pointer = ListGetNext(timer_pointer);
  if (!timer_pointer) {
    finish_playing();
    return;
  }

  if (ListGetAction(timer_pointer)) {
    event.type      = KeyPress;
    event.xkey.type = KeyPress;
  } else {
    event.type      = KeyRelease;
    event.xkey.type = KeyRelease;
  }
  event.xkey.send_event = True;
  event.xkey.display    = output_display;
  event.xkey.window     = output_window;
  event.xkey.root       = output_root_window;
  event.xkey.subwindow  = 0;
  event.xkey.state      = ListGetState(timer_pointer);
  event.xkey.keycode    = XKeysymToKeycode(output_display,
					   ListGetKey(timer_pointer));
  /* event.xkey.same_screen = False; */
  /* state ꤷʤƤ⡤ɥưʤСShift β
     ٥Ȥᤷư뤬
     state ꤷʤȡɥγ Shift 򲡤Ƥ饦ɥ
     äȤʤɤˡʸʸᤷʤ */

  /*
   * XSendEvent() 裳(propagate) True ˤƤʤȡ٥Ȥ
   * ¤ʤΤǡɥޥ͡ΥɥСʤɤ
   * ٥ȤäȤˡ¤줺˥٥ȤΤƤƤޤ
   */
  if (ListGetAction(timer_pointer)) {
    XSendEvent(output_display, output_window, True, KeyPressMask, &event);
    output_string("Key %s is pressed.",
		  XKeysymToString(ListGetKey(timer_pointer)));
  } else {
    XSendEvent(output_display, output_window, True, KeyReleaseMask, &event);
    output_string("Key %s is released.",
		  XKeysymToString(ListGetKey(timer_pointer)));
  }

  XFlush(output_display);

  if (ListGetNext(timer_pointer)) {
    t = cal_timeout(ListGetTimer(ListGetNext(timer_pointer)), play_speed);
    timeout_id = XtAddTimeOut(t, play_proc, NULL);
  } else { /* ǸޤǺäȤ */
    timer_pointer = NULL;
    finish_playing();
  }
}

/*---------------------------------------------------------------------------*/
/* 쥳ǥ󥰤ǡκ                                            */
/*---------------------------------------------------------------------------*/

static void play_sequence(XEvent * ev)
{
#if 0
  if (recording && pausing) recording = 0;
#endif

  if (playing) {
    playing++;
    output_message("Increased repeating times.");
    return;
  }

  if (pausing) {
    playing++;
    output_message("Now paused. Increased repeating times.");
    return;
  }

  if (timer_pointer) {
    if (ListGetNext(timer_pointer)) {
      playing = 1;
      timeout_id =
	XtAddTimeOut(cal_timeout(ListGetTimer(ListGetNext(timer_pointer)),
				 play_speed),
		     play_proc, NULL);
      output_message("Restart.");
    } else { /* ǸޤǺäȤ */
      timer_pointer = NULL;
      finish_playing();
    }
  } else {
    if (ListGetStart(key_list)) {
      playing = 1;
    }
    start_playing();
  }
}

/*---------------------------------------------------------------------------*/
/*                                                                   */
/*---------------------------------------------------------------------------*/

static void pause_sequence(XEvent * ev)
{
  if (!pausing) {
    pausing = 1;
    output_message("Paused. Select Pause menu again to restart.");
    if (timeout_id) {
      XtRemoveTimeOut(timeout_id);
      timeout_id = 0;
    }
    return;
  } else {
    pausing = 0;
    if (playing) restart_playing();
    else {
      output_message("Restart.");
    }
  }
}

/*---------------------------------------------------------------------------*/
/* ֤                                                              */
/*---------------------------------------------------------------------------*/

static void repeat_sequence(XEvent * ev)
{
  if (repeating) {
    repeating = 0;
    output_message("Auto repeat OFF.");
  } else {
    repeating = 1;
    output_message("Auto repeat ON.");
  }
}

/*---------------------------------------------------------------------------*/
/* ®٤®                                                        */
/*---------------------------------------------------------------------------*/

static void speed_up(XEvent * ev)
{
  if (play_speed > 0) {
    play_speed = play_speed - 1;
    output_value("Speed is %d/16.", speed[play_speed]);
  }
}

/*---------------------------------------------------------------------------*/
/* ®٤٤                                                        */
/*---------------------------------------------------------------------------*/

static void speed_down(XEvent * ev)
{
  if (play_speed < SPEED_NUM_MAX - 1) {
    play_speed = play_speed + 1;
    output_value("Speed is %d/16.", speed[play_speed]);
  }
}

/*---------------------------------------------------------------------------*/
/* ǡΥ                                                            */
/*---------------------------------------------------------------------------*/

static void save_sequence(XEvent * ev)
{
  FILE * fp;
  ListPointer p;

  p = ListGetStart(key_list);
  if (p == NULL) {
    output_message("No data.");
    return;
  }

  if (strcmp(filename, "-") == 0)
    fp = stdout;
  else
    fp = fopen(filename, "wt");

  if (fp == NULL) {
    output_string("Cannot open file %s.", filename);
    return;
  } else {
    output_message("Saving key sequence...");
    fprintf(fp, "# This is XKeyWrap key sequence data.\n");
    fprintf(fp, "# \n");
    fprintf(fp, "# Action(Push or Release) State Timer Key\n");
    while (p != NULL) {
      if (ListGetTimer(p) < 1000)
	fprintf(fp, "%c %d %3d %s\n",
		(ListGetAction(p) ? 'P' : 'R'),
		(int)ListGetState(p),
		(int)ListGetTimer(p),
		XKeysymToString(ListGetKey(p)));
      else
	fprintf(fp, "%c %d %d %s\n",
		(ListGetAction(p) ? 'P' : 'R'),
		(int)ListGetState(p),
		(int)ListGetTimer(p),
		XKeysymToString(ListGetKey(p)));
      p = ListGetNext(p);
    }
    fflush(fp);
    if (strcmp(filename, "-") == 0) {
      output_message("Saved to stdout, OK.");
    } else {
      fclose(fp);
      output_string("Saved to %s, OK.", filename);
    }
  }
}

/*---------------------------------------------------------------------------*/
/* ǡΥ                                                            */
/*---------------------------------------------------------------------------*/

/* եΥ꡼Ѵؿ */
void read_file(List key_list, char * filename)
{
  char buf[200];
  char * p;
  FILE * fp;
  char action;
  int state, timer;
  char key_string[200];

  if (filename == NULL) {
    output_message("Filename is not specified.");
    return;
  }

  if (key_list) ListDestroy(key_list);
  key_list = ListCreate();

  if (strcmp(filename, "-") == 0)
    fp = stdin;
  else
    fp = fopen(filename, "rt");

  if (fp == NULL) {
    output_string("Cannot open file %s.", filename);
    return;
  } else {
    while (fgets(buf, 200, fp) != NULL) {
      p = buf;
      while (*p == ' ' || *p == '\t') p++;

      if (*p != 'p' && *p != 'P' && *p != 'r' && *p != 'R')
	continue;

      action = '\0';
      state  = 0;
      timer  = 10;
      key_string[0] = '\0';
      sscanf(p, "%c %d %d %s", &action, &state, &timer, key_string);
      if (action != 'p' && action != 'P' && action != 'r' && action != 'R')
	continue;
      if (key_string[0] == '\0')
	continue;

      ListAddToEnd(key_list,
		   (action == 'p' || action == 'P') ? 1 : 0,
		   XStringToKeysym(key_string),
		   state, timer);
    }
    output_string("File %s was read. Select Play menu to play it.", filename);

    if (strcmp(filename, "-") != 0)
      fclose(fp);
  }
}

static void load_sequence(XEvent * ev)
{
  read_file(key_list, filename);
  output_message(NULL);
}

/*---------------------------------------------------------------------------*/
/* λ                                                                      */
/*---------------------------------------------------------------------------*/

static void quit(XEvent * ev)
{
  /* ҥץˤϡҥץλ */
  if (auto_quit) {
    if (child) {
      kill(child, SIGINT);
      child = 0;
    }
  }

  if (child_argv) free(child_argv);

  if (display) {
    if (font_gc) XFreeGC(display, font_gc);
    if (font) XUnloadFont(display, font);
  }
  if (output_display) {
    if (cursor_defined) {
      XUndefineCursor(output_display, output_root_window);
      cursor_defined = 0;
    }
    if (select_cursor) XFreeCursor(output_display, select_cursor);
  }

  if (key_list) ListDestroy(key_list);
  if (timeout_id) XtRemoveTimeOut(timeout_id);
  if (children_list) XFree(children_list);

  if (toplevel != NULL) XtDestroyWidget(toplevel);

  if (output_display != NULL && output_display != display) {
    XFlush(output_display);
    XCloseDisplay(output_display);
  }

  if (display != NULL) {
    XFlush(display);
    XCloseDisplay(display);
  }

  exit (0);
}

/*****************************************************************************/
/* ٥ȥϥɥ                                                          */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* ݡ                                                              */
/*---------------------------------------------------------------------------*/

static void expose(Widget w, XtPointer client_data, XEvent * ev,
		   Boolean * dispatch)
{
  output_message(NULL);
}

/*---------------------------------------------------------------------------*/
/* ꥵѤͽ                                                          */
/*---------------------------------------------------------------------------*/

/* This function is reserved for resize event. */
#if 0 /* ѤƤʤؿ -Wall ǥ˥󥰤ˤʤΤǡ#if 0 Ǥ */
static void resize(Widget w, XtPointer client_data, XEvent * ev,
		   Boolean * dispatch)
{
  XResizeWindow(display, XtWindow(toplevel),
		FONT_WIDTH * menubar_len, FONT_HEIGHT * 2);
  output_message(NULL);
}
#endif

/*---------------------------------------------------------------------------*/
/* ɥޥ͡㤫 f.delete ؤб                              */
/*---------------------------------------------------------------------------*/

static void wm_quit(Widget w, XEvent * ev, String * params, Cardinal * num)
{
  if ((ev->type == ClientMessage) && 
      (ev->xclient.data.l[0] != wm_delete_window)) {
    XBell(display, 0);
    XFlush(display);
  } else {
    quit(NULL);
  }
}

/*===========================================================================*/
/* βΥ                                                          */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* 򲡲                                                                */
/*---------------------------------------------------------------------------*/

static void key_press(Widget w, XtPointer client_data, XEvent * ev,
		      Boolean * dispatch)
{
  static XEvent event;
  KeySym key_sym;
  char * key_str;
  char string[20];

  if (ev->type == KeyPress) {
    event.type = KeyPress;
    event.xkey.type = KeyPress;
    event.xkey.send_event = True;
    event.xkey.display = output_display;
    event.xkey.window = output_window;
    event.xkey.root = output_root_window;
    event.xkey.subwindow = 0;
    event.xkey.state = ev->xkey.state;
    /* key_sym = XKeycodeToKeysym(display, ev->xkey.keycode, 0); */
    XLookupString(&(ev->xkey), string, 20, &key_sym, NULL);
    key_str = XKeysymToString(key_sym);
    event.xkey.keycode = XKeysymToKeycode(output_display, key_sym);
    /* event.xkey.same_screen = False; */
    /* XLookupString() ȡShift ʤɤβᤷƥ
       ΤǡShift + a  A ˤʤ */
    /* XLookupString() ˤʸȡShift ʤɤʸʤ */
    /* XKeycodeToKeysym() ȡa  Shift + a Ʊˤʤ */
    /* XKeysymToKeycode() ȡa  A Ʊɤˤʤ */

    if (output_window) { /* output_window && !playing ˤۤ⡩ */
      /*
       * XSendEvent() 裳(propagate) True ˤƤʤȡ٥Ȥ
       * ¤ʤΤǡɥޥ͡ΥɥСʤɤ
       * ٥ȤäȤˡ¤줺˥٥ȤΤƤƤޤ
       */
      XSendEvent(output_display, output_window, True, KeyPressMask, &event);
      XFlush(output_display);
      output_string("%s is pressed and sent to output window.", key_str);
    } else {
      output_string("%s is pressed.", key_str);
    }

    if (recording) {
      if (!pausing) {
	if (recording_timer)
	  timer_pointer = ListAdd(key_list, timer_pointer,
				  1, key_sym, ev->xkey.state,
				  ev->xkey.time - recording_timer);
	else
	  timer_pointer = ListAdd(key_list, timer_pointer,
				  1, key_sym, ev->xkey.state, 0);
      }
      recording_timer = ev->xkey.time;
    }
  }
}

/*---------------------------------------------------------------------------*/
/* Υ                                                              */
/*---------------------------------------------------------------------------*/

static void key_release(Widget w, XtPointer client_data, XEvent * ev,
			Boolean * dispatch)
{
  static XEvent event;
  KeySym key_sym;
  char * key_str;
  char string[20];

  if (ev->type == KeyRelease) {
    event.type = KeyRelease;
    event.xkey.type = KeyRelease;
    event.xkey.send_event = True;
    event.xkey.display = output_display;
    event.xkey.window = output_window;
    event.xkey.root = output_root_window;
    event.xkey.subwindow = 0;
    event.xkey.state = ev->xkey.state;
    /* key_sym = XKeycodeToKeysym(display, ev->xkey.keycode, 0); */
    XLookupString(&(ev->xkey), string, 20, &key_sym, NULL);
    key_str = XKeysymToString(key_sym);
    event.xkey.keycode = XKeysymToKeycode(output_display, key_sym);
    /* event.xkey.same_screen = False; */

    if (output_window) { /* output_window && !playing ˤۤ⡩ */
      /*
       * XSendEvent() 裳(propagate) True ˤƤʤȡ٥Ȥ
       * ¤ʤΤǡɥޥ͡ΥɥСʤɤ
       * ٥ȤäȤˡ¤줺˥٥ȤΤƤƤޤ
       */
      XSendEvent(output_display, output_window, True, KeyReleaseMask, &event);
      XFlush(output_display);
      output_string("%s is released and sent to output window.", key_str);
    } else {
      output_string("%s is released.", key_str);
    }

    if (recording) {
      if (!pausing) {
	if (recording_timer)
	  timer_pointer = ListAdd(key_list, timer_pointer,
				  0, key_sym, ev->xkey.state,
				  ev->xkey.time - recording_timer);
	else
	  timer_pointer = ListAdd(key_list, timer_pointer,
				  0, key_sym, ev->xkey.state, 0);
      }
      recording_timer = ev->xkey.time;
    }
  }
}

/*****************************************************************************/
/* ʥ                                                              */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* ҥץλȤΥʥ(SIGCHLD)Υϥɥ                     */
/*---------------------------------------------------------------------------*/

static void child_exit(int value)
{
  output_message("Child process was killed.");
  child = 0;
  output_window = 0;

  if (auto_quit) quit(NULL);
}

/*---------------------------------------------------------------------------*/
/* Ctrl-C                                                                    */
/*---------------------------------------------------------------------------*/

static void break_exit(int value)
{
  quit(NULL);
}

/*****************************************************************************/
/* ץν                                                          */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* إפν                                                              */
/*---------------------------------------------------------------------------*/

void output_help()
{
  fprintf(stdout, "%s", help_message);
  exit (0);
}

/*---------------------------------------------------------------------------*/
/* README ν                                                             */
/*---------------------------------------------------------------------------*/

void output_readme()
{
  fprintf(stdout, "%s", readme);
  exit (0);
}

/*---------------------------------------------------------------------------*/
/* COPYRIGHT ν                                                          */
/*---------------------------------------------------------------------------*/

void output_copyright()
{
  fprintf(stdout, "%s", copyright);
  exit (0);
}

/*---------------------------------------------------------------------------*/
/* פˤʤäޥɥ饤󥪥ץκ                                */
/*---------------------------------------------------------------------------*/

static void delete_argument(int * argcp, char ** argv, int n)
{
  if (n < *argcp) {
    for (; n < *argcp - 1; n++) {
      argv[n] = argv[n + 1];
    }
    argv[n] = NULL;
    (*argcp)--;
  }
}

/*---------------------------------------------------------------------------*/
/* ץθ                                                          */
/*---------------------------------------------------------------------------*/

int search_argument(int argc, char * argv[], char * opt)
{
  int i;
  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], opt)) return (i);
  return (-1);
}

/*---------------------------------------------------------------------------*/
/* ץβ                                                          */
/*---------------------------------------------------------------------------*/

void read_arguments(int * argcp, char ** argv)
{
  int i, n;

  /* -x ץν */
  /* -x ץϡֺǽ˽ʤФʤʤ         */
  /* (-x θˡ¾Υץ󤬸Ƥޤǽ뤿) */
  if ((n = search_argument(*argcp, argv, "-x")) > 0) {
    child_argv = (char **)malloc(sizeof(char *) * (*argcp - n));
    for (i = 0; i < *argcp - n - 1; i++) {
      child_argv[i] = argv[i + n + 1];
    }
    child_argv[i] = NULL;

    *argcp = n;
    argv[n] = NULL;

    if (child_argv[0] == NULL) {
      free(child_argv);
      child_argv = NULL;
    }
  }

  /* -h, -help ץν */
  if ((n = search_argument(*argcp, argv, "-h"   )) > 0) output_help();
  if ((n = search_argument(*argcp, argv, "-help")) > 0) output_help();

  /* -readme ץν */
  if ((n = search_argument(*argcp, argv, "-readme")) > 0) output_readme();

  /* -copyright ץν */
  if ((n = search_argument(*argcp, argv, "-copyright")) > 0)
    output_copyright();

  /* -p ץν */
  if ((n = search_argument(*argcp, argv, "-p")) > 0) {
    delete_argument(argcp, argv, n);
    playing = 1;
  }

  /* -play ץν */
  if ((n = search_argument(*argcp, argv, "-play")) > 0) {
    delete_argument(argcp, argv, n);
    playing = 1;
  }

  /* -r ץν */
  if ((n = search_argument(*argcp, argv, "-r")) > 0) {
    delete_argument(argcp, argv, n);
    repeating = 1;
  }

  /* -repeat ץν */
  if ((n = search_argument(*argcp, argv, "-repeat")) > 0) {
    delete_argument(argcp, argv, n);
    repeating = 1;
  }

  /* -rec ץν */
  if ((n = search_argument(*argcp, argv, "-rec")) > 0) {
    delete_argument(argcp, argv, n);
    recording = 1;
  }

  /* -record ץν */
  if ((n = search_argument(*argcp, argv, "-record")) > 0) {
    delete_argument(argcp, argv, n);
    recording = 1;
  }

  /* -pause ץν */
  if ((n = search_argument(*argcp, argv, "-pause")) > 0) {
    delete_argument(argcp, argv, n);
    pausing = 1;
  }

  /* -q ץν */
  if ((n = search_argument(*argcp, argv, "-q")) > 0) {
    delete_argument(argcp, argv, n);
    auto_quit = 1;
  }

  /* -quit ץν */
  if ((n = search_argument(*argcp, argv, "-quit")) > 0) {
    delete_argument(argcp, argv, n);
    auto_quit = 1;
  }

  /* -s ץν */
  if ((n = search_argument(*argcp, argv, "-s")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      play_speed = atoi(argv[n]);
      delete_argument(argcp, argv, n);
      if (play_speed < 0                ) play_speed = 0;
      if (play_speed > SPEED_NUM_MAX - 1) play_speed = SPEED_NUM_MAX - 1;
    }
  }

  /* -speed ץν */
  if ((n = search_argument(*argcp, argv, "-speed")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      play_speed = atoi(argv[n]);
      delete_argument(argcp, argv, n);
      if (play_speed < 0                ) play_speed = 0;
      if (play_speed > SPEED_NUM_MAX - 1) play_speed = SPEED_NUM_MAX - 1;
    }
  }

  /* -f ץν */
  if ((n = search_argument(*argcp, argv, "-f")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      filename = argv[n];
      delete_argument(argcp, argv, n);
    }
  }

  /* -file ץν */
  if ((n = search_argument(*argcp, argv, "-file")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      filename = argv[n];
      delete_argument(argcp, argv, n);
    }
  }

  /* -output-display ץν */
  if ((n = search_argument(*argcp, argv, "-output-display")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      output_display = XOpenDisplay(argv[n]);
      if (output_display == NULL) quit(NULL);
      output_root_window = DefaultRootWindow(output_display);
      delete_argument(argcp, argv, n);
    }
  }

  /* -output-window ץν */
  if ((n = search_argument(*argcp, argv, "-output-window")) > 0) {
    delete_argument(argcp, argv, n);
    if (n < *argcp) {
      output_window = atoi(argv[n]);
      delete_argument(argcp, argv, n);
    }
  }
}

/* ˥塼С */
void make_menubar()
{
  int i;

  menubar[0] = '\0';
  for (i = 0; menu_struct[i].name; i++) {
    strncat(menubar, menu_struct[i].name, 200);
    strncat(menubar, " ", 200);
  }
  menubar_len = strlen(menubar) - 1;
  menubar[menubar_len] = '\0';
}

/* ҥץǤʤä */
static void child_exec_error(int value)
{
  fprintf(stderr, "Child process exited.\n");
  exit (1);
}

void make_child(char ** child_argv)
{
  Window r, p;
  Window * children_before;
  Window * children_after;
  unsigned int n_before;
  unsigned int n_after;
  time_t t, t_old;

  if (child_argv == NULL) {
    return;
  } else if (child_argv[0] == NULL) {
    return;
  } else {
    if (output_window == 0) {
      XQueryTree(output_display, output_root_window, &r, &p,
		 &children_before, &n_before);
    }

    signal(SIGCHLD, child_exec_error);

    if ((child = fork()) == 0) {
      execvp(child_argv[0], child_argv);
      fprintf(stderr, "Can't exec %s.\n", child_argv[0]);
      exit (1);
    }

    if (output_window == 0) {
      time(&t_old);
      do {
	time(&t);
	if (t - t_old > 30) break; /* ॢ */
	XQueryTree(output_display, output_root_window,
		   &r, &p, &children_after, &n_after);
	output_window = diff_windows(children_before, n_before,
				     children_after, n_after);
	XFree(children_after);
	usleep(100);
      } while (!output_window);
      XFree(children_before);
    }
  }
}

/*****************************************************************************/
/* ᥤ                                                                    */
/*****************************************************************************/

int main(int argc, char * argv[])
{
  int i;

  /* ץν */
  read_arguments(&argc, argv);

  /* ˥塼С */
  make_menubar();

  /* ȥåץ٥ΥåȤ */
  toplevel = XtInitialize(argv[0], "XKeyWrap", NULL, 0, &argc, argv);
  if (toplevel == NULL) quit(NULL);
  XtVaSetValues(toplevel, XtNinput, True, NULL);
  /* XtVaSetValues(toplevel, XtNtitle, "XKeyWrap", NULL); */
  XtVaSetValues(toplevel, XtNwidth    , FONT_WIDTH * menubar_len , NULL);
  XtVaSetValues(toplevel, XtNminWidth , FONT_WIDTH * menubar_len , NULL);
  XtVaSetValues(toplevel, XtNmaxWidth , FONT_WIDTH * menubar_len , NULL);
  XtVaSetValues(toplevel, XtNheight   , FONT_HEIGHT * 2, NULL);
  XtVaSetValues(toplevel, XtNminHeight, FONT_HEIGHT * 2, NULL);
  XtVaSetValues(toplevel, XtNmaxHeight, FONT_HEIGHT * 2, NULL);

  /* ̣Υץ󤬤ä顤إפ */
  if (argc > 1) {
    for (i = 1; i < argc; i++)
      fprintf(stderr, "Unknown option : %s\n", argv[i]);
    output_help();
  }

  /* ǥץ쥤 */
  display = XtDisplay(toplevel);
  root_window = DefaultRootWindow(display);
  if (output_display == NULL) {
    output_display = display;
    output_root_window = root_window;
  }

  /* ҥץ */
  if (child_argv) make_child(child_argv);

  /* եȤγ */
  font = XLoadFont(display, "8x16");
  font_gc = XCreateGC(display, root_window, 0, 0);
  XSetFont(display, font_gc, font);
  XSetFunction(display, font_gc, GXcopy);

  /*  */
  select_cursor = XCreateFontCursor(output_display, XC_target);

  /*  */
  XtAddActions(actions, XtNumber(actions));

  /* ȥ󥹥졼 */
  XtOverrideTranslations(toplevel, XtParseTranslationTable(translations));

  /* ٥ȥϥɥϿ */
  XtAddEventHandler(toplevel, ExposureMask, False, expose, NULL);
  XtAddEventHandler(toplevel, KeyPressMask, False, key_press, NULL);
  XtAddEventHandler(toplevel, KeyReleaseMask, False, key_release, NULL);
  XtAddEventHandler(toplevel, ButtonPressMask, False, button_press, NULL);
  XtAddEventHandler(toplevel, PointerMotionMask, False, pointer_motion, NULL);

  /* åȤΥꥢ饤 */
  XtRealizeWidget(toplevel);
  while (!XtIsRealized(toplevel)) { /* None */ }

  /* ɥޥ͡㤫 f.delete ؤб */
  wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(display, XtWindow(toplevel), &wm_delete_window, 1);

  /* 襦ɥɽ */
  if (output_window)
    output_value("Window ID %d is selected.", (int)output_window);
  else
    output_message("No window is selected.");

  /* 󥹤ΥꥹȤν */
  key_list = ListCreate();

  /* եΥ꡼ */
  if (filename) read_file(key_list, filename);
  else filename = DEFAULT_FILENAME;

  /* ư˰ */
  if (pausing) {
    pausing = 0;
    pause_sequence(NULL);
  }

  /* ư˥ȤǷ֤ */
  if (repeating) {
    repeating = 0;
    repeat_sequence(NULL);
  }

  /* ư˥ȤǵϿ */
  if (recording) {
    recording = 0;
    record_sequence(NULL);
  }

  /* ư˥ȤǺ */
  if (playing) {
    playing = 0;
    play_sequence(NULL);
  }

  /* ʥ */
  signal(SIGCHLD, child_exit);
  signal(SIGINT , break_exit);
  signal(SIGTERM, break_exit);

  /* ᥤ롼 */
  XtMainLoop();

  return (0);
}

/*****************************************************************************/
/* End of Program.                                                           */
/*****************************************************************************/
